part of tecfy_core_client_package;

class DatabaseService {
  static Database _db;
  static String pathString;
  static String _dbPath = 'main.db';
  static LocalStorage storage;
  static bool isInit = false;
  static bool isEntityTableCreated = false;
  static bool get isLocalDbEnabled => _db != null;
  static bool get isSupportLocalDB =>
      Platform.isAndroid || Platform.isIOS || Platform.isMacOS;
  static init() async {
    try {
      if (kIsWeb ?? false) {
        storage = new LocalStorage('localDB.json');
      } else if (isSupportLocalDB) {
        pathString =
            '${(await path.getApplicationDocumentsDirectory()).path}${Platform.isWindows ? '\\' : '/'}';
        print('Database Path $pathString   $_dbPath');
        //await deleteDatabase(_dbPath);
        _db = await openDatabase(
            Platform.isWindows ? '$pathString$_dbPath' : _dbPath,
            version: 2,
            onCreate: (db, v) => _createDatabase(db, 0, v),
            onUpgrade: _createDatabase,
            onDowngrade: _createDatabase);
        print('db has been inited successfully !!!!!!!!!!!!!');
      } else {
        pathString = "c:\\data\\";
      }
    } catch (err) {
      // cannot get path means it's windows
      pathString = "c:\\data\\";
      print('xxxxxxxxxxxxxx  error inital database xxxxxxxxxxxx');
      print(err);
    }
    isInit = true;
  }

  static Future<bool> exeBatch(Future<bool> Function(Batch) process,
      [bool continueOnError = false]) async {
    try {
      if (_db == null) return false;
      var batch = _db.batch();
      if (await process(batch)) {
        batch.commit(continueOnError: continueOnError);
      }
      return true;
    } catch (err) {
      return false;
    }
  }

  static Future<List<Map<String, dynamic>>> rawQuery(String sql,
      [List<dynamic> arguments]) {
    if (_db == null || !_db.isOpen) {
      return Future.value([]);
    }
    return _db.rawQuery(sql, arguments);
  }

  static Future<int> insert(String table, Map<String, dynamic> values,
      {String nullColumnHack, ConflictAlgorithm conflictAlgorithm}) async {
    if (_db == null || !_db.isOpen) {
      print(
          'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX   db is not oppened!');
      return 0;
    }
    await _db.insert(table, values,
        nullColumnHack: nullColumnHack, conflictAlgorithm: conflictAlgorithm);
    return 10;
  }

  static Future<int> delete(String table,
      {String where, List<dynamic> whereArgs}) {
    if (_db == null || !_db.isOpen) {
      return Future.value(0);
    }
    return _db.delete(table, where: where, whereArgs: whereArgs);
  }

  static Future<int> update(String table, Map<String, dynamic> values,
      {String where,
      List<dynamic> whereArgs,
      ConflictAlgorithm conflictAlgorithm}) {
    if (_db == null || !_db.isOpen) {
      return Future.value(0);
    }
    return _db.update(table, values,
        where: where,
        whereArgs: whereArgs,
        conflictAlgorithm: conflictAlgorithm);
  }

  static Future<List<Map<String, dynamic>>> query(String table,
      {bool distinct,
      List<String> columns,
      String where,
      List<dynamic> whereArgs,
      String groupBy,
      String having,
      String orderBy,
      int limit,
      int offset}) {
    if (_db == null || !_db.isOpen) {
      return Future.value([]);
    }
    return _db.query(table,
        distinct: distinct,
        columns: columns,
        where: where,
        whereArgs: whereArgs,
        groupBy: groupBy,
        limit: limit,
        orderBy: orderBy,
        offset: offset);
  }

  static Future<void> _createDatabase(
      Database _db, int _fromVerison, int _toVersion) async {
    try {
      // var fmVersion =0;
      // var appVersion =0;

      if (_fromVerison <= 1) {
        await _db.execute(
            'CREATE TABLE Message (id TEXT PRIMARY KEY, fromUserId TEXT,toUserId TEXT,body TEXT, fromName TEXT,toName TEXT, time INTEGER,imSender INTEGER)');
        await _db.execute(
            'CREATE TABLE Notification (id TEXT PRIMARY KEY, titleAr TEXT,titleEn TEXT,messageAr TEXT, messageEn TEXT, time INTEGER, seen INTEGER, type INTEGER, payload TEXT)');
        await _db.execute(
            'CREATE TABLE Entity (id TEXT PRIMARY KEY, entity TEXT,body TEXT, timeStamp INTEGER, recordNumber INTEGER,isPosted INTEGER,postMessage TEXT)');
        await _db.execute(
            'CREATE TABLE DeleteLog (id TEXT PRIMARY KEY, entity TEXT, timeStamp INTEGER,isPosted INTEGER,postMessage TEXT)');

        await _db.execute(
            'CREATE INDEX INX_Entity_entity ON Entity(entity,timeStamp)');
        await _db
            .execute('CREATE INDEX INX_Entity_isPosted ON Entity(isPosted)');
        await _db.execute(
            'CREATE INDEX INX_DeleteLog_isPosted ON DeleteLog(isPosted)');

        await _createDatabaseEntityTable(_db);
      }
      await TecfyCoreApp.updateSqlLiteDb(_db, _fromVerison, _toVersion);
    } catch (err) {
      print(err);
    }
  }

  static List<String> builtInFields = [
    'id',
    'isPosted',
    'postMessage',
    'timeStamp'
  ];

  static Future<void> _createDatabaseEntityTable(Database _db) async {
    try {
      for (var entity in EntityService.entities.values) {
        List<String> cols = [
          'id TEXT PRIMARY KEY',
          'isPosted INTEGER',
          'postMessage TEXT',
          'timeStamp INTEGER'
        ];

        for (var prop in entity.properties) {
          if (builtInFields.contains(prop.dbName)) {
            continue;
          }
          if (prop.propertyType == PropertyTypes.File) {
            cols.add('${prop.dbName}Id TEXT');
            continue;
          }
          var type = 'TEXT';
          switch (prop.valueType) {
            case ValueTypes.DateTime:
              type = 'INTEGER';
              break;
            case ValueTypes.Boolean:
              type = 'INTEGER';
              break;
            case ValueTypes.Number:
              type = 'INTEGER';
              break;
            // case ValueTypes.NumberArray:
            //   type = 'TEXT';
            //   break;
            default:
          }
          cols.add('${prop.dbName} $type');
        }
        await _db.execute(
            'CREATE TABLE IF NOT EXISTS ${entity.dbName} (${cols.join(', ')})');
      }
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX _createDatabaseEntityTable');
      print(err);
    }
  }

  static Future<void> deleteDatabaseFile() async {
    try {
      await deleteDatabase(_dbPath);
    } catch (err) {
      print('xxxxxxxxxxxxxx  error deleting database xxxxxxxxxxxx');
      print(err);
    }
  }

  static Future<bool> saveEntity(String entityName, List<dynamic> data,
      {bool isPosted = true, bool overwriteLocal = false}) async {
    if (entityName.startsWith('_')) {
      var old = (await readCollection(entityName)).data ?? [];
      old.removeWhere((e) => data.any((d) => e['id'] == d['id']));
      old.addAll(data);
      return (await writeCollection(entityName, old)).success;
    }
    if (!ConnectionService.isOfflineMode) return false;
    // print(
    //     'Start ----------------------- entity save $entityName ${data.length}');

    var ids = data.map<String>((d) => d['id']).join("','");

    var delResult = await DatabaseService.delete(entityName,
        where: "${overwriteLocal ? '' : 'isPosted = 1 and '} id in ('$ids')");

    var result = await addEntityAll(entityName, data, isPosted: isPosted);
    return result;
  }

  static Future<bool> addEntityAll(String entityName, List<dynamic> data,
      {bool isPosted = true}) async {
    if (!ConnectionService.isOfflineMode) return false;
    try {
      // print(
      //     'Start ----------------------- entity add all $entityName ${data.length}');
      var entity = EntityService.getEntity(entityName);

      List<Map<String, dynamic>> dataList = [];
      for (var d in data) {
        var newData = toDatabaseEntity(entity, d);
        newData['isPosted'] = isPosted ? 1 : 0;
        dataList.add(newData);
      }
      var result = await DatabaseService.exeBatch((batch) async {
        for (var item in dataList) {
          batch.insert(entityName, item);
        }
        return true;
      });
      // print(
      //     '---------------- entity saved $entityName ${dataList.length} $result');
      return result;
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX addEntityAll');
      print(err);
      return false;
    }
  }

  static Future<bool> saveEntitySingle(String entityName, Map d,
      {bool isPosted = true}) async {
    if (d == null) return false;
    if (entityName.startsWith('_')) {
      var old = (await readCollection(entityName)).data ?? [];
      old.removeWhere((e) => e['id'] == d['id']);
      old.add(d);
      return (await writeCollection(entityName, old)).success;
    }

    if (!ConnectionService.isOfflineMode) return false;
    // print('Start ----------------------- entity single save $entityName ');
    var entity = EntityService.getEntity(entityName);
    var newData = toDatabaseEntity(entity, d);
    newData['isPosted'] = isPosted ? 1 : 0;
    var isExistResult = await DatabaseService.rawQuery(
        'select count(*) from $entityName where id = ?', [d['id']]);
    bool isNew = false;
    if (isExistResult == null ||
        isExistResult.length == 0 ||
        isExistResult.first.values.first == 0) {
      isNew = true;
    }
    int result;
    if (isNew)
      result = await DatabaseService.insert(entityName, newData);
    else
      result = await DatabaseService.update(entityName, newData,
          where: 'id = ?', whereArgs: [d['id']]);

    // print('---------------- entity single saved $entityName  $result');
    return result > 0;
  }

  static Future<List<dynamic>> getEntities(String entity,
      {String where = '',
      List<dynamic> whereArgs,
      String orderBy = 'timeStamp',
      int offset,
      int limit}) async {
    if (entity.startsWith('_')) {
      return (await readCollection('_ServerConnections')).data;
    }
    var entityStructure = EntityService.getEntity(entity);
    var items = (await DatabaseService.query(entity,
            where: where.isEmpty ? null : where,
            whereArgs: whereArgs,
            orderBy: orderBy,
            offset: offset,
            limit: limit))
        .map((d) => fromDatabaseEntity(entityStructure, d))
        .toList();
    print(
        '----------------------------- get $entity from local DB (${items.length})');

    return items;
  }

  static Future<dynamic> getEntityById(String entity, String id) async {
    if (!ConnectionService.isOfflineMode) return null;
    var items =
        (await DatabaseService.query(entity, where: 'id = ?', whereArgs: [id]))
            //.map((d) => jsonDecode(d['body']))
            .toList();

    return items.length > 0
        ? fromDatabaseEntity(EntityService.getEntity(entity), items[0])
        : null;
  }

  static Future<OperationResult> deleteEntityById(
      String entity, String id) async {
    if (entity.startsWith('_')) {
      var old = (await readCollection(entity)).data;
      old.removeWhere((e) => e['id'] == id);
      return (await writeCollection(entity, old));
    }
    if (!ConnectionService.isOfflineMode) return OperationResult(success: true);
    await DatabaseService.rawQuery('delete from DeleteLog where isPosted = 1');
    var addingResult = await DatabaseService.insert('DeleteLog', {
      'id': id,
      'timeStamp': DateTime.now().millisecondsSinceEpoch,
      'entity': entity,
      'postMessage': '',
      'isPosted': 0
    });
    if (addingResult > 0) {
      addingResult = await DatabaseService.delete(entity,
          where: 'id = ?', whereArgs: [id]);
    }

    // .exeBatch((b) async {
    //   b.insert('DeleteLog', {
    //     'id': id,
    //     'timeStamp': DateTime.now().millisecondsSinceEpoch,
    //     'entity': entity,
    //     'message': '',
    //     'isPosted': 0
    //   });
    //   b.delete('Entity', where: 'id = ?', whereArgs: [id]);
    //   await b.commit();
    //   return true;
    // });

    // var items = (await DatabaseService.delete('Entity',
    //     where: 'id = ?', whereArgs: [id]));

    return OperationResult(success: addingResult > 0);
  }

  static Future<OperationResult> deleteEntityBulk(List<String> ids) async {
    if (!ConnectionService.isOfflineMode) return OperationResult(success: true);
    await DatabaseService.exeBatch((b) async {
      for (var id in ids) {
        b.delete('Entity', where: 'id = ?', whereArgs: [id]);
      }
      await b.commit();
      return true;
    });
    return OperationResult(success: true);
  }

  // static Future<OperationResult> insert(
  //     String table, Map<String, dynamic> data) async {
  //   if (db == null || !db.isOpen) return OperationResult(success: false);
  //   var res = await db.insert(table, data);
  //   return OperationResult(success: res > 0);
  // }

  static Future<OperationResult> writeCollection(
      String entity, List<dynamic> data) async {
    try {
      if (storage != null) {
        storage.setItem(entity, data);
      } else {
        var file = File("$pathString$entity.json");
        await file.writeAsString(json.encode(data));
      }
      return OperationResult(success: true);
    } catch (err) {
      print('XXXXXXXXXXXXXXXXX  write local collection ($entity) $err');
      return OperationResult(success: false, message: err.toString());
    }
  }

  static Future<OperationResult<List<dynamic>>> readCollection(
      String entity) async {
    try {
      while (!isInit) await Future.delayed(Duration(milliseconds: 100));

      // print('--------------- read local collection ($entity)------------------');
      List<dynamic> data;
      if (kIsWeb) {
        data = storage.getItem(entity);
      } else {
        var file = File("$pathString$entity.json");
        data = json.decode(await file.readAsString());
      }
      if (data == null) return OperationResult(success: false);
      return OperationResult(success: true, data: data);
    } catch (err) {
      //print('XXXXXXXXXXXXXXXXX  read local collection ($entity) ("$pathString$entity.json") $err');
      return OperationResult(success: false, message: err.toString());
    }
  }

  static Map fromDatabaseEntity(Entity entity, Map databaseEntity) {
    var newData = Map<String, dynamic>();
    if (databaseEntity['id'] == null) {
      databaseEntity['id'] = Uuid().v4().replaceAll('-', '');
    }
    newData['id'] = databaseEntity['id'];

    if (databaseEntity['timeStamp'] != null) {
      if (databaseEntity['timeStamp'] is DateTime)
        newData['timeStamp'] =
            DateTime.fromMicrosecondsSinceEpoch(databaseEntity['timeStamp']);
    } else {
      newData['timeStamp'] = DateTime.now();
    }

    for (var prop in entity.properties) {
      if (prop.propertyType == PropertyTypes.File) {
        newData['${prop.dbName}Id'] = databaseEntity['${prop.dbName}Id'];
        continue;
      }
      if (databaseEntity[prop.dbName] == null ||
          builtInFields.contains(prop.dbName)) {
        continue;
      }
      switch (prop.valueType) {
        case ValueTypes.Boolean:
          newData[prop.dbName] = (databaseEntity[prop.dbName] == 1);
          break;
        case ValueTypes.DateTime:
          newData[prop.dbName] =
              DateTime.fromMillisecondsSinceEpoch(databaseEntity[prop.dbName]);
          break;
        case ValueTypes.Object:
          if (databaseEntity[prop.dbName] != null)
            newData[prop.dbName] = json.decode(databaseEntity[prop.dbName]);
          break;
        case ValueTypes.ObjectArray:
          if (databaseEntity[prop.dbName] != null)
            newData[prop.dbName] = json.decode(databaseEntity[prop.dbName]);
          break;
        case ValueTypes.Number:
          newData[prop.dbName] = databaseEntity[prop.dbName];
          break;
        default:
          newData[prop.dbName] = databaseEntity[prop.dbName].toString();
      }
    }
    return newData;
  }

  static Map toDatabaseEntity(Entity entity, Map entityData) {
    if (entity == null || entityData == null) return null;
    var newData = Map<String, dynamic>();
    if (entityData['id'] == null) {
      entityData['id'] = Uuid().v4().replaceAll('-', '');
    }
    newData['id'] = entityData['id'];

    if (entityData['timeStamp'] != null) {
      if (entityData['timeStamp'] is DateTime)
        newData['timeStamp'] =
            (entityData['timeStamp'] as DateTime).millisecondsSinceEpoch;
      else {
        var date = BaseModel.parseUTC(entityData['timeStamp'].toString());
        if (date != null) newData['timeStamp'] = date.millisecondsSinceEpoch;
      }
    } else {
      newData['timeStamp'] = DateTime.now().millisecondsSinceEpoch;
    }

    for (var prop in entity.properties) {
      if (prop.propertyType == PropertyTypes.File) {
        newData['${prop.dbName}Id'] = entityData['${prop.dbName}Id'];
        continue;
      }
      if (builtInFields.contains(prop.dbName)) {
        continue;
      }
      switch (prop.valueType) {
        case ValueTypes.Boolean:
          newData[prop.dbName] = entityData[prop.dbName] == true ? 1 : 0;
          break;
        case ValueTypes.DateTime:
          if (entityData[prop.dbName] is DateTime)
            newData[prop.dbName] =
                (entityData[prop.dbName] as DateTime).millisecondsSinceEpoch;
          else {
            var date = BaseModel.parseUTC(entityData[prop.dbName].toString());
            if (date != null)
              newData[prop.dbName] = date.millisecondsSinceEpoch;
          }
          break;
        case ValueTypes.Object:
          if (entityData[prop.dbName] != null) {
            newData[prop.dbName] = json.encode(entityData[prop.dbName]);
          }
          break;
        case ValueTypes.ObjectArray:
          if (entityData[prop.dbName] != null) {
            newData[prop.dbName] = json.encode(entityData[prop.dbName]);
          }
          break;
        case ValueTypes.Number:
          newData[prop.dbName] = entityData[prop.dbName];
          break;
        default:
          if (entityData[prop.dbName] != null)
            newData[prop.dbName] = entityData[prop.dbName].toString();
      }
    }
    return newData;
  }
}
