part of tecfy_core_client_package;

class InLineForm extends StatefulWidget {
  final String entityId;
  final Entity entity;
  final dynamic data;
  final bool inPopup;
  final bool readOnly;
  final void Function(dynamic) onDismiss;
  final void Function(dynamic) onChanged;
  InLineForm(this.entityId,
      {this.entity,
      this.data,
      this.onChanged,
      this.onDismiss,
      this.inPopup = false,
      this.readOnly = false});
  @override
  _InLineFormState createState() => _InLineFormState();
}

class _InLineFormState extends State<InLineForm> {
  List<Property> props;
  List<Property> calcProps;
  List<List<Property>> lines;
  // Parser parser;
  @override
  void initState() {
    props = widget.entity.properties.where((d) => d.isEdit).toList();
    lines = [[]];
    for (var p in props) {
      lines.last.add(p);
      if (p.newLineAfter) lines.add([]);
    }
    calcProps = props
        .where((p) =>
            p.propertyType == PropertyTypes.Number &&
            (p as PropertyNumber).calcFormula != null &&
            (p as PropertyNumber).calcFormula.isNotEmpty)
        .toList();
    super.initState();
  }

  _onUpdateData(Map data, Property property) {
    // we need this for Ramy's application to update the records when select an Item
    if (property.propertyType == PropertyTypes.Dropdown) {
      var props = (property as PropertyDropdown).overwritePropertiesOnSelect;
      for (var key in data.keys) {
        if (props.contains(key)) onValueChanged(data[key], key);
      }
      setState(() {});
    }
  }

  onValueChanged(value, propName) async {
    print('inline value change! $propName');
    widget.data[propName] = value;

    bool updated = await widget.entity.setDefaultValues(widget.data);

    for (Property p in props) {
      var r = await p.isVisble(widget.data, widget.entity);
      if (r != p.isVisable) {
        p.isVisable = r;
        updated = true;
      }
    }

    // check the calced properties
    for (PropertyNumber p in calcProps) {
      if (p.calcedProps.contains(p.upperFirst(propName)) ||
          p.calcedProps.contains(propName)) {
        var vars = p.calcedProps
            .map((pro) => widget.data[p.lowerFirst(pro)] ?? 0)
            .toList();
        widget.data[p.dbName] = p.getCalcResult(vars);
        setState(() {});
      }
    }

    if (updated || props.any((p) => p.filterValueField == propName))
      setState(() {});

    widget?.onChanged(widget.data);
  }

  @override
  Widget build(BuildContext context) {
    return widget.onDismiss == null || widget.readOnly
        ? getData()
        : Stack(children: <Widget>[
            getData(),
            Align(
                alignment: LocalConfigService.isArabic
                    ? Alignment.topLeft
                    : Alignment.topRight,
                child: IconButton(
                    icon: Icon(Icons.close),
                    onPressed: () => widget.onDismiss(widget.data)))
          ]);
  }

  // get _getMargen => (widget.inPopup ?? false) ? 35 : 20;

  _getWidth(Property prop) {
    if (prop.fullWidth) return double.infinity;
    var fullwidth = LayoutService.getWidth(context);

    fullwidth -= (ThemeService.menuWidth + 100);
    if (fullwidth > ThemeService.containerWidth) {
      fullwidth = ThemeService.containerWidth - 50;
    }
    if (fullwidth < 600) {
      return double.infinity; // means take full width
    }
    return fullwidth / 2;
  }

  getData() {
    return SizedBox(
      width: MediaQuery.of(context).size.width -
          ((widget.inPopup ?? false) ? 35 : 20),
      child: Card(
        color: ThemeService.backgroundSecondColor,
        margin: EdgeInsets.all(10),
        child: false
            ? Container(
                margin: EdgeInsets.all(0),
                constraints: BoxConstraints(maxWidth: double.infinity),
                padding: EdgeInsets.all(10),
                child: Column(
                  children: lines
                      .map<Widget>(
                        (props) => Row(
                          children: List.generate(
                            props.length,
                            (i) => Expanded(
                              flex: props[i].size == PropertySizes.xSmall
                                  ? 1
                                  : props[i].size == PropertySizes.Small
                                      ? 2
                                      : props[i].size == PropertySizes.Default
                                          ? 3
                                          : props[i].size == PropertySizes.Large
                                              ? 4
                                              : props[i].size ==
                                                      PropertySizes.xLarge
                                                  ? 5
                                                  : 3,
                              child: PorpertyComponent(
                                readOnly: widget.readOnly,
                                property: props[i],
                                data: widget.data,
                                onChange: (v) =>
                                    onValueChanged(v, props[i].dbName),
                                onUpdateData: (v) => _onUpdateData(v, props[i]),
                              ),
                            ),
                          ),
                        ),
                      )
                      .toList(),
                ),
              )
            : Wrap(
                children: List.generate(
                  props.length,
                  (i) => Container(
                    margin: EdgeInsets.all(0),
                    constraints: BoxConstraints(maxWidth: _getWidth(props[i])),
                    padding: EdgeInsets.all(10),
                    child: PorpertyComponent(
                      readOnly: widget.readOnly,
                      property: props[i],
                      data: widget.data,
                      onChange: (v) => onValueChanged(v, props[i].dbName),
                      onUpdateData: (v) => _onUpdateData(v, props[i]),
                    ),
                  ),
                ),
              ),
      ),
    );
  }
}
