part of tecfy_core_client_package;

class PorpertyComponent extends StatefulWidget {
  final Property property;
  final Entity entity;
  final Color valueColor;
  final void Function(dynamic) onChange;
  final void Function(Map) onUpdateData;
  //final dynamic initalValue;
  /// if this property is a dropdown and it's filterd by another field
  //final dynamic filterValue;
  final Map<String, dynamic> data;
  final bool readOnly;
  PorpertyComponent(
      {this.property,
      this.entity,
      this.onChange,
      this.onUpdateData,
      this.valueColor,
      //this.filterValue,
      this.data,
      this.readOnly = false});
  @override
  _PorpertyComponentState createState() => _PorpertyComponentState();
}

class _PorpertyComponentState extends State<PorpertyComponent> {
  TextEditingController textController;
  FocusNode focusNode;
  dynamic _initValue;
  //bool isVisable =true;

  @override
  void initState() {
    init();
    super.initState();
  }

  init() {
    _initValue = widget.data[widget.property.dbName];
    var val = _initValue?.toString();
    if (widget.property.propertyType == PropertyTypes.Number && val != null) {
      if (val.contains('.'))
        val = (_initValue as num).toDouble().toStringAsFixed(2);
    }
    if (widget.property.propertyType == PropertyTypes.Text ||
        widget.property.propertyType == PropertyTypes.Number) {
      if (textController == null) {
        textController = TextEditingController(text: val ?? '');
        focusNode = FocusNode();
        textController.addListener(() {
          submitChages();
        });
      } else {
        textController.text = val ?? '';
      }
    }
  }

  submitChages() {
    if (widget.onChange == null) return;
    if (widget.property.propertyType == PropertyTypes.Text) {
      _initValue = textController.text;
      widget.onChange(textController.text);
    }
    if (widget.property.propertyType == PropertyTypes.Number) {
      _initValue = num.tryParse(textController.text);
      widget.onChange(_initValue);
    }
  }

  @override
  Widget build(BuildContext context) {
    if (!widget.property.isVisable) return Container();

    if (_initValue != widget.data[widget.property.dbName]) init();

    if (widget.readOnly) return getReadOnlyWidget();

    if (widget.property.readOnly &&
        widget.property.propertyType != PropertyTypes.Dropdown &&
        widget.property.propertyType != PropertyTypes.OtherEntityMany &&
        widget.property.propertyType != PropertyTypes.ManyToMany &&
        widget.property.propertyType != PropertyTypes.OneToMany)
      return TextFormLabel(
        label: widget.property.name,
        value: _initValue?.toString() ?? '__',
      );

    switch (widget.property.propertyType) {
      case PropertyTypes.Dropdown:
        return DropdownAdvancedEntity(
            (widget.property as PropertyDropdown).entityName,
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            entity: EntityService.getEntity(
                (widget.property as PropertyDropdown).entityName),
            url: (widget.property as PropertyDropdown).dropdownUrl,
            property: (widget.property as PropertyDropdown),
            data: widget.data, onChanged: (v, data) {
          widget.onChange(v);
          if (data != null && widget.onUpdateData != null)
            widget.onUpdateData(data);
        });
      case PropertyTypes.OtherEntityMany:
        return ListFormProperty(
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            entityId: (widget.property as PropertyOtherEntityMany).entityName,
            data: _initValue,
            onChanged: (v) {
              widget.onChange(v);
            });

      case PropertyTypes.ManyToMany:
        if (!(widget.property as PropertyManyToMany).isWithData)
          return ManyToManyProperty(
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            selectedValues: _initValue,
            entity: EntityService.getEntity(
                (widget.property as PropertyManyToMany).entityName),
            onChanged: widget.onChange,
          );

        return ManyToManyWithData(
          (widget.property as PropertyManyToMany).entityName,
          label: widget.property.name,
          isRequired: widget.property.isRequired,
          items: _initValue,
          entity: EntityService.getEntity(
              (widget.property as PropertyManyToMany).entityName),
          onChanged: widget.onChange,
          entityMiddle:
              (widget.property as PropertyManyToMany).middleEntityName,
          propertieMiddle:
              (widget.property as PropertyManyToMany).middleOtherPropertyName,
        );

      case PropertyTypes.OneToMany:
        return ManyToManyProperty(
          label: widget.property.name,
          isRequired: widget.property.isRequired,
          selectedValues: _initValue,
          entity: EntityService.getEntity(
              (widget.property as PropertyOneToMany).entityName),
          onChanged: widget.onChange,
        );

      case PropertyTypes.Text:
        if ((widget.property as PropertyText).isBarcode) {
          return BarcodePropertyComponenet(
            labelText: widget.property.name,
            isRequired: widget.property.isRequired,
            textStyle: widget.valueColor == null
                ? null
                : TextStyle(color: widget.valueColor),
            focusNode: focusNode,
            controller: textController,
            onFieldSubmitted: (v) => focusNode?.unfocus(),
            onEditingComplete: submitChages,
          );
        }
        return AppTextFormField(
          textDirection: LocalConfigService.isArabic &&
                  !(widget.property as PropertyText).isEmail
              ? TextDirection.rtl
              : TextDirection.ltr,
          labelText: widget.property.name,
          isRequired: widget.property.isRequired,
          textStyle: widget.valueColor == null
              ? null
              : TextStyle(color: widget.valueColor),
          focusNode: focusNode,
          controller: textController,
          onFieldSubmitted: (v) => focusNode?.unfocus(),
          textInputAction: TextInputAction.next,
          obscureText: (widget.property as PropertyText).isPassword,
          keyboardType: (widget.property as PropertyText).isEmail
              ? TextInputType.emailAddress
              : (widget.property as PropertyText).lines > 1
                  ? TextInputType.multiline
                  : TextInputType.text,
          maxLines: (widget.property as PropertyText).lines,
          onEditingComplete: submitChages,
        );
      case PropertyTypes.Number:
        return AppTextFormField(
            labelText: widget.property.name,
            isRequired: widget.property.isRequired,
            controller: textController,
            focusNode: focusNode,
            onFieldSubmitted: (v) => focusNode?.unfocus(),
            textStyle: widget.valueColor == null
                ? null
                : TextStyle(color: widget.valueColor),
            textInputAction: TextInputAction.next,
            keyboardType: TextInputType.number,
            onEditingComplete: submitChages);
      case PropertyTypes.Boolean:
        return AppCheckBox(
            label: widget.property.name,
            initalValue: _initValue,
            onChanged: (v) {
              widget.onChange(v);
            });
      case PropertyTypes.File:
        return PhotoProperty(
            fileData: _initValue,
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            fileId: widget.data[widget.property.dbName + 'Id'],
            onChanged: (v) => widget.onChange(v));
      case PropertyTypes.Color:
        return ColorPropertyComponenet(
            initalValue: _initValue,
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            onChanged: (v) => widget.onChange(v));
      case PropertyTypes.MapArea:
        return MapAreaPropertyComponenet(
            initalValue: _initValue,
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            onChanged: (v) => widget.onChange(v));
      case PropertyTypes.MarkImage:
        return MarkImagePropertyComponenet(
            initalValue: _initValue,
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            onChanged: (v) => widget.onChange(v));
      case PropertyTypes.Date:
        return DateProperty(
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            initalValue: _initValue,
            onChanged: (v) {
              widget.onChange(v);
            });
      case PropertyTypes.Location:
        return LocationPropertyComponenet(
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            initalValue: _initValue,
            onChanged: (v) {
              widget.onChange(v);
            });
      case PropertyTypes.MapArea:
        return MapAreaPropertyComponenet(
            label: widget.property.name,
            isRequired: widget.property.isRequired,
            initalValue: _initValue,
            onChanged: (v) {
              widget.onChange(v);
            });
      default:
        return Text(
            '${widget.property.dbName} ${widget.property.propertyType}');
    }
  }

  Widget getReadOnlyWidget() {
    switch (widget.property.propertyType) {
      case PropertyTypes.Dropdown:
        var e = EntityService.lowerFirst(
            (widget.property as PropertyDropdown).entityName);
        return TextFormLabel(
          label: widget.property.name,
          value: widget.data[e] != null
              ? (LocalConfigService.isArabic
                      ? widget.data[e]['nameAr']
                      : widget.data[e]['nameEn']) ??
                  widget.data[e]['display'] ??
                  widget.data[e]['name'] ??
                  widget.data[e]['title'] ??
                  '__'
              : '__',
        );
      case PropertyTypes.OtherEntityMany:
        print(_initValue);
        return ListFormProperty(
            readOnly: true,
            label: widget.property.name,
            entityId: (widget.property as PropertyOtherEntityMany).entityName,
            data: _initValue,
            onChanged: (v) {});
      default:
        try {
          return TextFormWidget(
            label: widget.property.name,
            value: widget.property.getDisplayWidget(widget.data),
          );
        } catch (err) {
          return TextFormLabel(
            label: widget.property.name,
            value: _initValue?.toString() ?? '__',
          );
        }
    }
  }
}
